/*=============================================================================
gdmalloc - Memory leak detection helper utility
gdmalloc.c - CTOR hook
  Copyright (C) Jan "Hirogen2" Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  http://linux01.org:2222/f/hxtools.tbz2
===============================================================================
It can do:
- find unfreed pointers
- find double free actions
- fence-pre and fence-post write detection
At the moment, it does not do anymore, as these are the main leaks that hide in
applications.

GDmalloc is intended to be as free as possible. You do not need to change any
of your code to get it working. Just pass -lgdmalloc to the linker, that is it.

It provides good stack backtrace info even if no -g option was specified,
however what you WILL need is a good symbol table. However (2), all GDmalloc
can find out ATM is the function, not the exact line. (At least not yet)
However (3), you can put a printf() statement before your
call to malloc() to identify it as the one (see the 'F' flag too!).

You will need a Glibc with symbol table, otherwise dynamically loaded .so files
via dlopen() will not be resolved. I recommend to compile a Glibc in debug mode
for this (it is about 19MB for me, in /usr/lib/libc_nostrip.so.6), however,
that will take its time.

And thus, GDmalloc only works under Linux, with GCC/GNU ld and Glibc. God, all
those underscores to keep namespace clean. Oh well. This was nearly as hard as
writing segvtracer.


Options are set via the environment variable GDMALLOC. Current flags:
  A     activate GDmalloc (so you can have a persistent -lgdmalloc)
  C     do not use ANSI color
  F     activate flush for stdout, stderr (note that GDmalloc is ran before
        any of your code!)
  f     Activate fence-write detection
  v     print malloc and free calls as well as allocated object count


-----------------------------------------------------------------------------*/
#define _GNU_SOURCE 1
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <malloc.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libHX.h>

typedef unsigned char byte;
typedef unsigned long Ulong;

#define MAXFNLEN 1024
#define MAXLNLEN 1024
#define SBT_SIZE 256 /* maximum buffer size for stack backtrace dumps */
#define _weak_alias(__name, __aname) \
  extern __typeof(__name) __aname __attribute__((weak, alias(#__name)));
#define _strong_alias(__name, __aname) \
  extern __typeof(__name) __aname __attribute__((alias(#__name)));

enum {
  FENCE_PRE  = 1 << 0,
  FENCE_POST = 1 << 1,
};

// Function prototypes
extern void *__libc_malloc(size_t);
extern void *__libc_calloc(size_t, size_t);
extern void *__libc_realloc(void *, size_t);
extern void __libc_free(void *);

extern void _GDM_init(void *) __attribute__((constructor));
extern void _GDM_deinit(void) __attribute__((destructor));
extern void *_GDM_malloc(size_t);
extern void *_GDM_calloc(size_t, size_t);
extern void *_GDM_realloc(void *, size_t);
extern void _GDM_free(void *);
extern void *_GDM_malloc_hook(size_t, const void *);
extern void *_GDM_realloc_hook(void *, size_t, const void *);
extern void _GDM_free_hook(void *, const void *);

static int _GDM_check_dlopen(void);
static int _GDM_check_fence_P(const void *);
static void _GDM_check_fence(void);
static void _GDM_dump_unfreed(void);
static int _GDM_flag(char);
static const void *_GDM_get_ebp(int);
static void _GDM_hooks_deploy(void);
static void _GDM_hooks_restore(void);
static int _GDM_load_file(const char *, unsigned long);
static int _GDM_print(char, char *, ...);
static const char *_GDM_rsymbol(const void *);
static char *_GDM_stack_backtrace(void);
static void _GDM_update_maps(void);
static int _GDM_value_cmp(void *, void *);

// Variables
static void *(*ORIG_hmalloc)(size_t, const void *) = NULL;
static void *(*ORIG_hrealloc)(void *, size_t, const void *) = NULL;
static void (*ORIG_hfree)(void *, const void *) = NULL;
static pthread_mutex_t ext_lock = PTHREAD_MUTEX_INITIALIZER;
static struct HXbtree *_GDM_aobj = NULL, *_GDM_map = NULL, *_GDM_stab = NULL;
static byte _GDM_sid[16];
static const void *_GDM_BP = NULL;
static char *flags = NULL;
static int _GDM_inited = 0, _GDM_fence_cnt = 0;

//-----------------------------------------------------------------------------
void _GDM_init(void *caller) {
    void *rd;

    if(_GDM_inited) { return; }
    pthread_mutex_init(&ext_lock, NULL);
    pthread_mutex_lock(&ext_lock);
    if(_GDM_inited++) { return; }

    flags = getenv("GDTRAP");
    if(flags != NULL) { _GDM_BP = (const void *)strtoul(flags, NULL, 0); }

    flags = getenv("GDMALLOC");
    if(!_GDM_flag('A')) { return; }

    rd = HX_rand_init(HXRAND_DEV_URANDOM | HXRAND_SYS_RAND);
    if(rd != NULL) {
        byte *sp = _GDM_sid, i = 16;
        while(i--) { *sp++ = HX_rand(rd); }
        HX_rand_deinit(rd);
    } else {
        memcpy(_GDM_sid, "GDMALLOC2003", 16);
    }

    _GDM_aobj = HXbtree_init(HXBT_HASH | HXBT_LKEY | HXBT_PCMP, _GDM_value_cmp);
    _GDM_map = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcmp);
    _GDM_stab = HXbtree_init(HXBT_HASH | HXBT_LKEY | HXBT_CPTR | HXBT_PCMP, _GDM_value_cmp);

    ORIG_hmalloc = __malloc_hook;
    ORIG_hrealloc = __realloc_hook;
    ORIG_hfree = __free_hook;

    if(_GDM_flag('F')) {
        setvbuf(stdout, NULL, _IONBF, 0);
        setvbuf(stderr, NULL, _IONBF, 0);
    }

    _GDM_update_maps();
    _GDM_print(0, "GDmalloc ready\n");
    pthread_mutex_unlock(&ext_lock);
    _GDM_hooks_deploy();
    return;
}

void _GDM_deinit(void) {
    if(!_GDM_flag('A')) { return; }
    _GDM_hooks_restore();
    _GDM_print(0, "GDmalloc hooks deinitialized\n");
    if(_GDM_aobj->itemcount > 0) { _GDM_dump_unfreed(); }
    HXbtree_free(_GDM_aobj);
    HXbtree_free(_GDM_map);
    HXbtree_free(_GDM_stab);
    pthread_mutex_destroy(&ext_lock);
    return;
}

//-----------------------------------------------------------------------------
// Re-wrappers
void *_GDM_malloc(size_t s) {
    return __libc_malloc(s);
}

void *_GDM_calloc(size_t n, size_t s) {
    return __libc_calloc(n, s);
}

void *_GDM_realloc(void *ptr, size_t s) {
    return __libc_realloc(ptr, s);
}

void _GDM_free(void *ptr) {
    __libc_free(ptr);
    return;
}

_strong_alias(_GDM_malloc, malloc);
_strong_alias(_GDM_calloc, calloc);
_strong_alias(_GDM_realloc, realloc);
_strong_alias(_GDM_free, free);

//-----------------------------------------------------------------------------
// Libc hooks
void *_GDM_malloc_hook(size_t s, const void *caller) {
    void *realp, *retp;
    size_t rs;

    pthread_mutex_lock(&ext_lock);
    _GDM_hooks_restore();
    if(_GDM_check_dlopen()) { _GDM_update_maps(); }

    rs = s + 5 * sizeof(void *);
    realp = __libc_malloc(rs);
    retp = NULL;

    if(realp != NULL) {
        /* Fence post write check
        [SID] [BLOCKLEN] [SID] [SID] [...data...] [SID]
        */
        if(_GDM_flag('f')) {
            memcpy(realp, _GDM_sid, sizeof(void *));
            memcpy(realp + sizeof(void *), &s, sizeof(void *));
            memcpy(realp + 2 * sizeof(void *), _GDM_sid, sizeof(void *));
            memcpy(realp + 3 * sizeof(void *), _GDM_sid, sizeof(void *));
            memcpy(realp + 4 * sizeof(void *) + s, _GDM_sid, sizeof(void *));
        }
        retp = realp + 4 * sizeof(void *);
        HXbtree_add(_GDM_aobj, realp, _GDM_stack_backtrace());
    }

    _GDM_print('v', "[objs=%5ld] 0x%08lX = malloc(s=%lu)\n",
     _GDM_aobj->itemcount, (Ulong)retp, (Ulong)s);

    if(_GDM_BP != NULL && retp == _GDM_BP) {
        // Make GDB stop
        volatile int *p = NULL;
        ++*p;
    }

    _GDM_hooks_deploy();
    pthread_mutex_unlock(&ext_lock);
    return retp;
}

void *_GDM_realloc_hook(void *iretp, size_t is, const void *caller) {
    void *irealp = NULL, *orealp, *oretp = NULL;
    size_t rs;

    pthread_mutex_lock(&ext_lock);
    _GDM_hooks_restore();

    rs = is + 5 * sizeof(void *);
    if(iretp != NULL) { irealp = iretp - 4 * sizeof(void *); }
    orealp = __libc_realloc(irealp, rs);

    if(orealp != irealp) {
        if(irealp != NULL) { HXbtree_del(_GDM_aobj, irealp); }
        if(orealp != NULL) { HXbtree_add(_GDM_aobj, orealp, _GDM_stack_backtrace()); }
    }
    if(orealp != NULL) {
        memcpy(orealp, _GDM_sid, sizeof(void *));
        memcpy(orealp + sizeof(void *), &is, sizeof(void *));
        memcpy(orealp + 2 * sizeof(void *), _GDM_sid, sizeof(void *));
        memcpy(orealp + 3 * sizeof(void *), _GDM_sid, sizeof(void *));
        memcpy(orealp + 4 * sizeof(void *) + is, _GDM_sid, sizeof(void *));
        oretp = orealp + 4 * sizeof(void *);
    }

    _GDM_print('v', "[objs=%5ld] 0x%08lX = realloc(ptr=0x%08lX, s=%lu)\n",
     _GDM_aobj->itemcount, (Ulong)oretp, (Ulong)iretp, (Ulong)is);

    _GDM_hooks_deploy();
    pthread_mutex_unlock(&ext_lock);
    return oretp;
}

void _GDM_free_hook(void *iretp, const void *caller) {
    void *cp, *irealp = NULL;
    int fp = 0;

    pthread_mutex_lock(&ext_lock);
    _GDM_hooks_restore();

    if(iretp != NULL) { irealp = iretp - 4 * sizeof(void *); }
    if(_GDM_flag('A') && irealp != NULL && (fp = _GDM_check_fence_P(irealp))) {
        if(fp & FENCE_PRE) {
            _GDM_print('f', "free> Fence/pre  broken for 0x%08lX\n",
             (Ulong)iretp);
        }
        if(fp & FENCE_POST) {
            _GDM_print('f', "free> Fence/post broken for 0x%08lX\n",
             (Ulong)iretp);
        }
    }
    __libc_free(irealp);

    _GDM_print('v', "[objs=%5ld] free(ptr=0x%08lX)\n",
     _GDM_aobj->itemcount, (Ulong)iretp);

    if((cp = HXbtree_del(_GDM_aobj, irealp)) == NULL && errno == ENOENT) {
        _GDM_print(0, "Tried to double-free object 0x%08lX\n", (Ulong)iretp);
    }

    if(_GDM_flag('A') && ++_GDM_fence_cnt % 100 == 0) { _GDM_check_fence(); }

    _GDM_hooks_deploy();
    pthread_mutex_unlock(&ext_lock);
    return;
}

//-----------------------------------------------------------------------------
// Internal functions
static int _GDM_check_dlopen(void) {
    const void *ebp = _GDM_get_ebp(3);
    int found = 0;

    while(ebp != 0x0) {
        const char *sym = _GDM_rsymbol(*(const void **)(ebp + sizeof(void *)));
        if(sym != NULL && strncmp(sym, "dlopen@@", 8) == 0) { ++found; break; }
        ebp = *(const void **)ebp;
    }

    return found;
}

static int _GDM_check_fence_P(const void *addr) {
    size_t s;
    int ret = 0;
    memcpy(&s, addr + sizeof(void *), sizeof(void *));

    if(memcmp(addr, _GDM_sid, sizeof(void *)) != 0 ||
     memcmp(addr + 2 * sizeof(void *), _GDM_sid, sizeof(void *)) != 0 ||
     memcmp(addr + 3 * sizeof(void *), _GDM_sid, sizeof(void *)) != 0) {
        ret |= FENCE_PRE;
    }

    if(memcmp(addr + 4 * sizeof(void *) + s, _GDM_sid, sizeof(void *)) != 0) {
        ret |= FENCE_POST;
    }
    return ret;
}

static void _GDM_check_fence(void) {
    struct HXbtrav *travp = HXbtrav_init(_GDM_aobj, NULL);
    struct HXbtree_node *n;
    int fp = 0;

    while((n = HXbtraverse(travp)) != NULL) {
        if((fp = _GDM_check_fence_P(n->key))) {
            if(fp & FENCE_PRE) {
                _GDM_print(0, "check_fence> Fence/pre  broken at 0x%08lX\n",
                 (Ulong)n->key);
            }
            if(fp & FENCE_POST) {
                _GDM_print(0, "check_fence> Fence/post broken at 0x%08lX\n",
                 (Ulong)n->key);
            }
        }
    }

    HXbtrav_free(travp);
    return;
}

static void _GDM_dump_unfreed(void) {
    struct HXbtrav *travp = HXbtrav_init(_GDM_aobj, NULL);
    struct HXbtree_node *n;

    while((n = HXbtraverse(travp)) != NULL) {
        _GDM_print(0, "Unfreed object 0x%08lX (Trace: %s)\n",
         (Ulong)n->key, n->data);
    }

    HXbtrav_free(travp);
    return;
}

static int _GDM_flag(char f) {
    if(f == '\0' || (flags != NULL && strchr(flags, f) != NULL)) { return 1; }
    return 0;
}

static const void *_GDM_get_ebp(int skip) {
    const void *ebp;
    asm volatile("movl %%ebp, %0;" :"=m"(ebp));
    while(skip--) {
        if(ebp == NULL) { return NULL; }
        ebp = *(const void **)ebp;
    }
    return ebp;
}

static void _GDM_hooks_deploy(void) {
    __malloc_hook = _GDM_malloc_hook;
    __realloc_hook = _GDM_realloc_hook;
    __free_hook = _GDM_free_hook;
    return;
}

static void _GDM_hooks_restore(void) {
    __malloc_hook = ORIG_hmalloc;
    __realloc_hook = ORIG_hrealloc;
    __free_hook = ORIG_hfree;
    return;
}

static int _GDM_load_file(const char *filename, unsigned long base) {
    char buf[MAXLNLEN];
    int c = 0;
    FILE *fp;

    snprintf(buf, MAXLNLEN, "nm %s 2>/dev/null", filename);
    if((fp = popen(buf, "r")) == NULL) { return 0; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *addrp = buf, *type, *name;
        unsigned long addr;

        if(*addrp == ' ') { continue; }
        if((type = strchr(buf, ' ')) == NULL) { continue; }
        *type++ = '\0';
        if(*type != 't' && *type != 'T') { continue; }
        if((name = strchr(type, ' ')) == NULL) { continue; }
        *name++ = '\0';

        HX_chomp(name);
        addr = strtoul(addrp, NULL, 16);
        if(addr < base) { addr += base; }
        HXbtree_add(_GDM_stab, (void *)addr, name);
        ++c;
    }

    _GDM_print('v', "Added %d symbols from %s\n", c, filename);
    pclose(fp);
    return 1;
}

static int _GDM_print(char f, char *fmt, ...) {
    int rv = 0;
    va_list argp;
    va_start(argp, fmt);
    if(_GDM_flag(f)) {
        if(!_GDM_flag('C') && isatty(fileno(stderr))) {
            fprintf(stderr, "\e[35mGDM\e[1m>\e[0m "); // ]]]
        } else {
            fprintf(stderr, "GDM> ");
        }
        rv = vfprintf(stderr, fmt, argp);
    }
    va_end(argp);
    return rv;
}

static const char *_GDM_rsymbol(const void *addr) {
    struct HXbtree *btree = _GDM_stab;
    struct HXbtree_node *node = btree->tree;
    char *sym = NULL;
    void *bp = NULL;
    int v;

    while(node != NULL) {
        v = btree->pcmp((void *)addr, node->key);
        if(node->key > bp && node->key <= addr) {
          sym = node->data;
          bp = node->key;
        }
        if(v == 0) { break; }
        node = node->st[v > 0];
    }

    return sym;
}

static char *_GDM_stack_backtrace(void) {
    // Skip 2(3) stack frames: _GDM_malloc, _GDM_malloc_hook (and _GDM_get_ebp)
    const void *ebp = _GDM_get_ebp(3);
    char buf[SBT_SIZE], ibuf[256];
    memset(buf, '\0', SBT_SIZE);

    while(ebp != 0x0) {
      const void *f_addr = *(const void **)(ebp + sizeof(void *));
      const char *sym = _GDM_rsymbol(f_addr);

      /*_GDM_print(0, "EBP=0x%08lX, f_addr=0x%08lX, sym=%s\n",
       (Ulong)ebp, (Ulong)f_addr, sym);*/

      if(sym == NULL) {
          snprintf(ibuf, 256, "0x%08lX", (Ulong)f_addr);
      } else {
          snprintf(ibuf, 256, "%s", sym);
      }

      HX_strlcat(buf, ibuf, SBT_SIZE);
      ebp = *(const void **)ebp;
      if(ebp != 0x0) { HX_strlcat(buf, " < ", SBT_SIZE); }
    }

    return HX_strdup(buf);
}

static void _GDM_update_maps(void) {
    FILE *fp;
    char buf[MAXLNLEN], fbuf[MAXFNLEN];
    snprintf(fbuf, MAXFNLEN, "/proc/%d/maps", getpid());
    if((fp = fopen(fbuf, "r")) == NULL) { return; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        char *base = buf;
        char *filename = strchr(buf, '/');
        if(filename == NULL) { continue; }
        HX_chomp(buf);
        *strchr(buf, '-') = '\0';
        if(HXbtree_find(_GDM_map, filename) == NULL) {
            _GDM_load_file(filename, strtoul(base, NULL, 16));
            HXbtree_add(_GDM_map, filename);
        }
    }

    fclose(fp);
    return;
}

static int _GDM_value_cmp(void *pa, void *pb) {
    return (pa > pb) - (pa < pb);
}

//==[ End of file ]============================================================
