/*=============================================================================
vitalnix/neluder - User Management UI
data.c - Data processing
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <crypt.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <libHX.h>
#include "neluder-src/global.h"

static char *transform7(byte *, char *, size_t);

static struct stab { char *set, *out; } subst_tab[] = {
  {"", "ae"}, {"", "oe"}, {"", "ue"}, {"", "ss"},
  {"", "a"}, {"", "c"}, {"", "e"}, {"", "i"},
  {"", "o"}, {"", "u"}, {"", "l"}, {"", "m"}, {"", "n"},
  {"", "r"}, {"", "y"}, {NULL, NULL},
};

//-----------------------------------------------------------------------------
char *date_scramble(char *input) {
  /* Scramble a date. The scrambling format is stored in the first four bits of
  the resulting string. */
  const byte fmt = 0;
  unsigned short day = 0, month = 0, year = 0;
  char *ret;

  if(input != NULL) {
    if(strchr(input, '.') != NULL) { // DD.MM.YYYY
     sscanf(input, "%hu.%hu.%hu", &day, &month, &year); }
    else if(strchr(input, '/') != NULL) { // MM/DD/YYYY
     sscanf(input, "%hu/%hu/%hu", &month, &day, &year); }
    else if(strchr(input, '-') != NULL) { // YYYY-MM-DD
     sscanf(input, "%hu-%hu-%hu", &year, &month, &day); }

    { // two- and three-year digit fixup (03, 103 => 2003)
      struct tm result;
      time_t now = time(NULL);
      int bp, nc;
      localtime_r(&now, &result);
      bp = (result.tm_year + 50) % 100;
      nc = result.tm_year - result.tm_year % 100 + ((bp < 50) ? 100 : 0);
      if(year >= 100 && year < 1000) { year += 1900; }
      else if(year < 100) { year += 1900 + (year > bp) ? (nc - 100) : nc; }
    }
  }

  ret = malloc(16);
  snprintf(ret, 16, "%08lX", ((unsigned long)fmt << 24) |
   ((unsigned long)year << 16) | ((unsigned long)month << 8) | day);
  return ret;
}

char *date_unscramble(char *input, char *output, size_t len, char sep) {
  unsigned short day, month, year;
  byte fmt = (*input & 0xF0) >> 4;

  if(fmt != 0) { return NULL; } // unsupported scrambling version
  sscanf(input, "%04hX%02hX%02hX", &year, &month, &day);

  if(sep == '.') { // DD.MM.YYYY
   snprintf(output, len, "%02hu.%02hu.%04hu", day, month, year); }
  else if(sep == '/') { // MM/DD/YYYY
   snprintf(output, len, "%02hu/%02hu/%04hu", month, day, year); }
  else if(sep == '-') { // YYYY-MM-DD
   snprintf(output, len, "%04hu-%02hu-%02hu", year, month, day); }
  return output;
}

char *fixup_login_name(char *ln, unsigned int idx, char *dest, size_t dsize) {
  /* Login names are fixup'ed by adding a unique number them, necessarily
  truncating part of the original login name. */
  char idxstr[16], tln[16];
  snprintf(idxstr, 16, "%u", idx);
  snprintf(tln, 8 - strlen(idxstr) + 1, "%s", ln);
  snprintf(dest, dsize, "%s%s", tln, idxstr);
  return dest;
}

char *make_login_name(char *in_nname, char *in_vname) {
  /* This takes the name (in_nname) and prenom (in_vname) and makes a 7-bit
  clean login name in the style VNNNNNN. Numerical indexes are done in
  fixup_login_name() (called from the main code). */
  char buf[8], fix_nname[8], fix_vname[8], *p;

  memset(fix_nname, '\0', 8);
  memset(fix_vname, '\0', 8);
  transform7(in_vname, fix_vname, 8);

  if(in_nname != NULL) {
    if((p = strchr(in_nname, ' ')) != NULL) { in_nname = ++p; }
    transform7(in_nname, fix_nname, 8);
    snprintf(buf, 8, "%c%s", *fix_vname, fix_nname);
  } else {
    snprintf(buf, 8, "%s", fix_vname);
  }

  return HX_strdup(HX_strlower(buf));
}

int only_digits(char *p) {
  while(*p) {
    if(!isdigit(*p++)) { return 0; }
  }
  return 1;
}

//-----------------------------------------------------------------------------
static char *transform7(byte *src, char *dest, size_t dsize) {
  /* Transform a name (possibly with umlauts, etc.) into a 7-bit clean,
  A-to-Z-only login name. */
  dest[--dsize] = '\0';
  while(dsize && *src != '\0') {
    if((*src >= 'A' && *src <= 'Z') || (*src >= 'a' && *src <= 'z')) {
      *dest++ = *src;
      --dsize;
    } else if(*src >= 128) {
      /* This loop may take "a while" (at least it is faster than something in
      Perl in any case). But it is worth it. Embedding regular expression
      substitution would be more of a task than this. */
      struct stab *sp = subst_tab;
      while(sp->set != NULL) {
        if(strchr(sp->set, *src) != NULL) {
          size_t l;
          if(dsize >= (l = strlen(sp->out))) {
            strncpy(dest, sp->out, l);
            dest += l;
            dsize -= l;
          } else {
            /* Do not let "abcdea" through, because if there was too few
            space, the "" gets skipped but dsize is still big enough to let
            the "a" in. */
            dsize = 0;
          }
          break;
        }
        ++sp;
      }
    }
    ++src;
  }
  return dest;
}

//==[ End of file ]============================================================
